/**
 * Miranda ICQ contact list and events database to Ickle database converter.
 * $Id: miranda2ickle.cpp,v 1.2 2002/05/30 16:00:07 berk Exp $
 *
 * @file  miranda2ickle.cpp
 * @brief Main program.
 *
 * Copyright (c) 2002, Stanislav Karchebny  <berk@upnet.ru>
 * Distributed under BSD License. See file LICENSE for details.
 */

#define __m2i_VERSION "0.7.0"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include "types.h"
#include "miranda_db.h"
#include "errwarn.h"
#include "vf.h"
#include "Contact.h"
#include "options.h"
#include "iconvert.h"
#include "sstream_fix.h"

using namespace std;

void all_contact_settings(const DBContact *contact, Contact *output_contact);
void all_contact_events(const DBContact *contact, Contact *output_contact);

char help_head[]    = "usage: miranda2ickle [options] mirandadbfile [ickledbpath]\n"
                      "where options are:\n";
char help_tail[]    = "\n"
                      "if ickledbpath is not specified, ~/.ickle will be used.\n"
#if HAVE_ICONV_OPEN
                      "if charsets are not specified, no translation will be made;\n"
                      "if one of charsets is not specified, current locale will be used ($LANG).\n"
#endif
                      "\n";

static char *mirandadb = NULL;
static char *icklepath = NULL;
static FILE *mirandafd = NULL;
int show_help = 0;

iconvert *converter;
#if HAVE_ICONV_OPEN
char *recode_from = NULL;
char *recode_to = NULL;
#endif

int not_an_option_handler( char *param )
{
   static int which_file = 1;

   if(which_file == 1)
   {
      mirandadb = param;
      which_file++;
   }
   else if(which_file == 2)
   {
      icklepath = param;
      which_file++;
   }
   else error("extraneous file %s specified on command line\n", param);

   return 0;
}

#if HAVE_ICONV_OPEN
int opt_recode_from( char *cmd, char *param, int extra )
{
   recode_from = param;
   return 0;
}

int opt_recode_to( char *cmd, char *param, int extra )
{
   recode_to = param;
   return 0;
}
#endif

int opt_help( char *cmd, char *param, int extra )
{
   show_help = 1;
   return 0;
}

opt_option options[] =
{
#if HAVE_ICONV_OPEN
   { 'f', "from",               1, opt_recode_from,    0,   "input charset",  "charset" },
   { 't', "to",                 1, opt_recode_to,      0,   "output charset", "charset" },
#endif
   { 'h', "help",               0, opt_help,           0,   "show usage information"    },
};

/*
 * main(argc,argv)
 */
int main(int argc, char **argv)
{
   notice("miranda2ickle database converter - version " __m2i_VERSION "\n"
          "Copyright (c) 2002, Stanislav Karchebny  <berk@upnet.ru>\n");

   if( parse_cmdline( argc, argv, options, sizeof(options)/sizeof(options[0]) ) ) return 1;
   if(show_help || !mirandadb)
   {
      help_msg(help_head, help_tail, options, sizeof(options)/sizeof(options[0]));
      exit(0);
   }

   if( access(mirandadb,0) ) error("cannot read mirandadbfile %s\n", mirandadb);

   if(icklepath)
   {
      ostringstream ostr;
      ostr << icklepath << "/";
      icklepath = strdup(ostr.str().c_str());
      if( access(icklepath,0) ) error("cannot access ickledbpath %s\n", icklepath);
      ostr << "contacts/";
      if( access(ostr.str().c_str(),0) ) error("cannot access ickledb contacts @ %s\n", ostr.str().c_str());
   }
   else
   {
      ostringstream ostr;
      if( !getenv("HOME") ) error("cannot access home directory, $HOME not set\n");
      ostr << getenv("HOME") << "/.ickle/";
//      icklepath = new char [strlen(getenv("HOME")) + strlen("/.ickle/") + 1];
//      sprintf(icklepath, "%s/.ickle/", getenv("HOME"));
      icklepath = strdup(ostr.str().c_str());
      if( access(icklepath,0) ) error("cannot access ickledbpath %s\n", icklepath);
      ostr << "contacts/";
      if( access(ostr.str().c_str(),0) ) error("cannot access ickledb contacts @ %s\n", ostr.str().c_str());
   }

   Dnotice("using\n.. mirandadbfile %s\n.. ickledbpath %s\n", mirandadb, icklepath);

#if HAVE_ICONV_OPEN
   // set up translation parameters and init iconv
   if( recode_from || recode_to ) // translation needed
   {
      char *locale = getenv("LANG");
      if( !locale ) error("cannot recode without $LANG, specify both --from and --to\n");
      locale = strrchr(locale, '.');   
      if( !locale ) error("cannot recode without proper $LANG, specify both --from and --to\n");
      if( !recode_from ) recode_from = locale+1;
      if( !recode_to ) recode_to = locale+1;

      converter = new iconvert(recode_from, recode_to);
      Dnotice(".. iconv translation from %s to %s\n", recode_from, recode_to);
   }
   else converter = new iconvert(); // identity converter
#else
   converter = new iconvert(); // identity converter
#endif

   // open miranda db and read header
   mirandafd = fopen(mirandadb, "rb");
   if( !mirandafd ) error("cannot open mirandadbfile %s for reading\n", mirandadb);

   DBHeader header;
   vfread( mirandadb, __FUNCTION__, __LINE__, mirandafd, &header, sizeof(header) );

   if( !ValidateDBHeader(&header) ) error("mirandadbfile has corrupt header, probably not miranda db?\n");

   Dnotice("mirandadbfile verified\n"
           ".. file end offset: %08x\n"
           ".. file slack     : %u bytes\n"
           ".. # of contacts  : %u\n"
           ".. 1st contact @  : %08x\n"
           ".. user info @    : %08x\n",
           header.ofsFileEnd,
           header.slackSpace,
           header.contactCount,
           header.ofsFirstContact,
           header.ofsUser
   );

   // Start measuring time consumption
   timeval start, end;
   gettimeofday(&start, NULL);

   // Read db contacts, converting and outputting them to ickledbpath/contacts
   DBContact contact;
   DWORD contact_offset;
   DWORD contact_count = 1;

   contact_offset = header.ofsFirstContact;
   while(contact_offset)
   {
      notice("processing contact %u  \r", contact_count++);
      ReadContact( mirandadb, mirandafd, contact_offset, &contact );
      Dnotice("\nread contact from %08x\n"
              "... settings @    : %08x\n"
              "... # of events   : %u\n"
              "... 1st event @   : %08x\n"
              "... last event @  : %08x\n"
              "... 1st unread @  : %08x\n"
              "... 1st unread tm : %s", // \n added by ctime
              contact_offset,
              contact.ofsFirstSettings,
              contact.eventCount,
              contact.ofsFirstEvent,
              contact.ofsLastEvent,
              contact.ofsFirstUnreadEvent,
              ctime((const time_t *)&contact.timestampFirstUnread)
      );

      Contact *ct_data = new Contact(icklepath);

      // read this contact settings
      all_contact_settings(&contact, ct_data);

      // read this contact events
      all_contact_events(&contact, ct_data);

      // write contact data to ickle data files
      ct_data->write();
      delete ct_data;

      // proceed to next contact
      contact_offset = contact.ofsNext;
   }

   // Read and convert user settings, preserving existing settings (output to ickle.conf_from_mirandadbfile)
   notice("processing user contact          \r");
   ReadContact( mirandadb, mirandafd, header.ofsUser, &contact );
   Dnotice("\nread contact from %08x\n"
           "... settings @    : %08x\n"
           "... # of events   : %u\n"
           "... 1st event @   : %08x\n"
           "... last event @  : %08x\n"
           "... 1st unread @  : %08x\n"
           "... 1st unread tm : %s", // \n added by ctime
           contact_offset,
           contact.ofsFirstSettings,
           contact.eventCount,
           contact.ofsFirstEvent,
           contact.ofsLastEvent,
           contact.ofsFirstUnreadEvent,
           ctime((const time_t *)&contact.timestampFirstUnread)
   );

   Contact *ct_data = new Contact(icklepath, "self");
   all_contact_settings(&contact, ct_data);
   all_contact_events(&contact, ct_data);
   ct_data->write();
   delete ct_data;

   // Calculate time consumption
   double time;
   gettimeofday(&end, NULL);
   if( timercmp(&start, &end, >) ) time = (start.tv_sec - end.tv_sec) + (start.tv_usec - end.tv_usec) / 1000000.; // start is over end
   else                            time = (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) / 1000000.;

   fclose(mirandafd);
   delete converter;

   notice("\nConversion successfully completed in %6.3f seconds! Enjoy your Ickle!\n", time);

   return 0;
}

// the settings related to multiple modules may be present, read them one by one and
// recover necessary information.
void all_contact_settings(const DBContact *contact, Contact *out_contact)
{
   DBContactSettings settings;
   DWORD settings_offset;
   char *modname;

   settings_offset = contact->ofsFirstSettings;
   while(settings_offset)
   {
      ReadContactSettings( mirandadb, mirandafd, settings_offset, &settings );
      // Find out the module these settings are for
      modname = ReadModuleName(mirandadb, mirandafd, settings.ofsModuleName);

      Dnotice("read contact settings from %08x\n"
              ".... owner module       : %s\n"
              ".... owner modulename @ : %08x\n"
              ".... payload size       : %u bytes\n"
              ".... payload @          : %08x\n",
              settings_offset,
              modname,
              settings.ofsModuleName,
              settings.cbBlob,
              settings_offset + sizeof(settings)
      );

      // Read every single setting of this chunk
      DBVARIANT setting;
      DWORD advance;
      settings_offset += sizeof(settings);
      while(settings.cbBlob)
      {
         advance = ReadSingleContactSetting( mirandadb, mirandafd, settings_offset, &setting, &settings.cbBlob );
         if( setting.type != DBVT_DELETED )
         {
            Dnotice("read single contact setting from %08x\n"
                    "..... name   : %s\n"
                    "..... type   : %s\n"
                    "..... value  : ",
                    settings_offset,
                    setting.name,
                    ContactSettingType(&setting)
            );
            switch( setting.type )
            {
               case DBVT_BYTE:    Dnotice("%02x (%u) (%c)\n", setting.bVal, setting.bVal, setting.bVal > 31 ? setting.bVal : '?'); break;
               case DBVT_WORD:    Dnotice("%04x (%u)\n", setting.wVal, setting.wVal); break;
               case DBVT_DWORD:   Dnotice("%08x (%u)\n", setting.dVal, setting.dVal); break;
               case DBVT_ASCIIZ:  Dnotice("%s\n", setting.pszVal); break;
               case DBVT_BLOB:    Dnotice("%u bytes @%08x\n", setting.cbVal, setting.pbVal); break;
               default:           Dnotice("???\n"); break;
            }

            // add the value to the value storage
            // we push all shit into the contact and let it figure out which ones it needs
            out_contact->add(modname, &setting);
         }

         settings_offset += advance;
      }

      delete modname;
      settings_offset = settings.ofsNext;
   }
}

void all_contact_events(const DBContact *contact, Contact *out_contact)
{
   DBEvent event;
   PBYTE data;
   DWORD events_offset;
   char *modname;

   events_offset = contact->ofsFirstEvent;
   while(events_offset)
   {
      data = ReadContactEvent( mirandadb, mirandafd, events_offset, &event );
      // Find out the module this event is for
      modname = ReadModuleName(mirandadb, mirandafd, event.ofsModuleName);

      Dnotice("read contact event from %08x\n"
              ".... owner module       : %s\n"
              ".... owner modulename @ : %08x\n"
              ".... event type         : %s\n"
              ".... event flags        : %s, %s\n"
              ".... payload size       : %u bytes\n"
              ".... payload @          : %08x\n" 
              ".... payload as string  : %s\n", 
              events_offset,
              modname,
              event.ofsModuleName,
              ContactEventType(&event),
              event.flags&DBEF_SENT?"SENT":"RECEIVED",
              event.flags&DBEF_READ?"READ":"NOT READ",
              event.cbBlob,
              events_offset + sizeof(event),
              converter->iconv((char *)data).c_str()
      );

      out_contact->add_history(modname, &event, data); // pass unconverted, since Contact takes care of translation

      events_offset = event.ofsNext;
   }
}
