/**
 * Miranda ICQ contact list and events database to Ickle database converter.
 * $Id: miranda_db.cpp,v 1.2 2002/05/30 16:00:07 berk Exp $
 *
 * @file  miranda_db.cpp
 * @brief Miranda DB manipulation functions.
 *
 * Copyright (c) 2002, Stanislav Karchebny  <berk@upnet.ru>
 * Distributed under BSD License. See file LICENSE for details.
 */

#include <stdio.h>
#include <string.h>
#include "types.h"
#include "miranda_db.h"
#include "errwarn.h"
#include "vf.h"

struct DBSignature {
  char name[15];
  BYTE eof;
};
static struct DBSignature dbSignature={"Miranda ICQ DB",0x1A};

/* return 0 if header is invalid, 1 otherwise */
int ValidateDBHeader( DBHeader *header )
{
   if( memcmp(header->signature,&dbSignature,sizeof(header->signature)) ) return 0;
   if( header->version != DB_THIS_VERSION ) return 0;
   if( header->ofsUser == 0 ) return 0;
   return 1;
}

int ReadContact( char *filename, FILE *fp, DWORD offset, DBContact *contact )
{
   fseek(fp, offset, SEEK_SET);
   vfread(filename, __FUNCTION__, __LINE__, fp, contact, sizeof(DBContact));
   if( contact->signature != DBCONTACT_SIGNATURE )
      error("ReadContact: contact signature @%08x incorrect, database corrupt?\n", offset);
   return 1;
}

int ReadContactSettings( char *filename, FILE *fp, DWORD offset, DBContactSettings *settings )
{
   fseek(fp, offset, SEEK_SET);
   vfread(filename, __FUNCTION__, __LINE__, fp, settings, sizeof(DBContactSettings));
   if( settings->signature != DBCONTACTSETTINGS_SIGNATURE )
      error("ReadContactSettins: settings signature @%08x incorrect, database corrupt?\n", offset);
   return 1;
}

/*
 * read single setting from blob at offset, returns number of bytes read.
 * bytesleft is decreased by number of bytes read.
 */
DWORD ReadSingleContactSetting( char *filename, FILE *fp, DWORD offset, DBVARIANT *setting, DWORD *bytesleft )
{
   DWORD bytesread = 0;

   if(!bytesleft || !*bytesleft) return 0;

   BYTE cbName;
   WORD cbString;

   fseek(fp, offset, SEEK_SET);

   // get the name length
   vfread( filename, __FUNCTION__, __LINE__, fp, &cbName, 1 );
   if( !cbName )
   {
      setting->type = DBVT_DELETED;
      *bytesleft = 0;
      return 1;
   }

   // get the name itself
   setting->name = new char [cbName+1];
   vfread( filename, __FUNCTION__, __LINE__, fp, setting->name, cbName );
   setting->name[cbName] = '\0';

   // read setting type
   vfread( filename, __FUNCTION__, __LINE__, fp, &setting->type, 1 );

   // read setting data (depends on type)
   bytesread = 1 + cbName + 1;
   switch( setting->type )
   {
      case DBVT_DELETED:
         warning("deleted entry, abandoned whole record\n");
         *bytesleft = bytesread;
         break;

      case DBVT_BYTE:
         vfread( filename, __FUNCTION__, __LINE__, fp, &setting->bVal, 1 );
         bytesread++;
         break;

      case DBVT_WORD:
         vfread( filename, __FUNCTION__, __LINE__, fp, &setting->wVal, 2 );
         bytesread += 2;
         break;

      case DBVT_DWORD:
         vfread( filename, __FUNCTION__, __LINE__, fp, &setting->dVal, 4 );
         bytesread += 4;
         break;

      case DBVT_ASCIIZ:
         vfread( filename, __FUNCTION__, __LINE__, fp, &cbString, 2 );
         setting->pszVal = new char [cbString+1];
         if(cbString > 0)
            vfread( filename, __FUNCTION__, __LINE__, fp, setting->pszVal, cbString );
         setting->pszVal[cbString] = '\0';
         bytesread += 2 + cbString;
         break;

      case DBVT_BLOB:
         vfread( filename, __FUNCTION__, __LINE__, fp, &setting->cbVal, 2 );
         if(setting->cbVal > 0)
         {
            setting->pbVal = new BYTE [setting->cbVal];
            vfread( filename, __FUNCTION__, __LINE__, fp, setting->pbVal, setting->cbVal );
         }
         else
         {
            setting->pbVal = NULL;
         }
         bytesread += 2 + setting->cbVal;
         break;

      default:
         error("ReadSingleContactSetting: unknown data type %u @%08x\n", setting->type, offset);
   }

   if( bytesread > *bytesleft )
   {
      *bytesleft = bytesread;
      warning("ouch.. read more bytes than there's left to\n");
   }

   *bytesleft -= bytesread;
   return bytesread;
}

const char *ContactSettingType(DBVARIANT *setting)
{
   switch( setting->type )
   {
      case DBVT_DELETED: return "DELETED";
      case DBVT_BYTE:    return "BYTE";
      case DBVT_WORD:    return "WORD";
      case DBVT_DWORD:   return "DWORD";
      case DBVT_ASCIIZ:  return "ASCIIZ";
      case DBVT_BLOB:    return "BLOB";
      default:           return "Unknown";
   }
}

char *ReadModuleName( char *filename, FILE *fp, DWORD offset )
{
   DBModuleName modname;

   fseek(fp, offset, SEEK_SET);
   vfread(filename, __FUNCTION__, __LINE__, fp, &modname, sizeof(DBModuleName));
   if( modname.signature != DBMODULENAME_SIGNATURE )
      error("ReadModuleName: module name signature @%08x incorrect, database corrupt?\n", offset);

   char *name = new char [modname.cbName+1];
   vfread( filename, __FUNCTION__, __LINE__, fp, name, modname.cbName );
   name[modname.cbName] = '\0';

   return name;
}

BYTE *ReadContactEvent( char *filename, FILE *fp, DWORD offset, DBEvent *event )
{
   BYTE *blob = NULL;

   fseek(fp, offset, SEEK_SET);
   vfread(filename, __FUNCTION__, __LINE__, fp, event, sizeof(DBEvent));
   if( event->signature != DBEVENT_SIGNATURE )
      error("ReadContactEvent: event signature @%08x incorrect, database corrupt?\n", offset);

   if( event->cbBlob > 0 )
   {
      blob = new BYTE [event->cbBlob];
      vfread(filename, __FUNCTION__, __LINE__, fp, blob, event->cbBlob);
   }

   return blob;
}

const char *ContactEventType(DBEvent *event)
{
   switch( event->eventType )
   {
      case EVENTTYPE_MESSAGE:     return "MESSAGE";
      case EVENTTYPE_URL:         return "URL";
      case EVENTTYPE_ADDED:       return "YOU WERE ADDED";
      case EVENTTYPE_AUTHREQUEST: return "AUTH REQ";
      case EVENTTYPE_FILE:        return "FILE";
      default:                    return "Unknown";
   }
}
