/**
 * Miranda ICQ contact list and events database to Ickle database converter.
 * $Id: miranda_db.h,v 1.9 2002/05/30 16:00:07 berk Exp $
 *
 * @file  miranda_db.h
 * @brief Miranda DB manipulation functions.
 *
 * Copyright (c) 2002, Stanislav Karchebny  <berk@upnet.ru>
 * Distributed under BSD License. See file LICENSE for details.
 *
 * Miranda ICQ 0.1.0.0 database format. Shamelessly stolen from Miranda source code.
 * http://miranda-icq.sourceforge.net
 * Copyright (C) 2000-2  Richard Hughes, Roland Rabien & Tristan Van de Vreede
 * Distributed under GNU GPL.
 */

#ifndef __M2I_MIRANDA_DB_H
#define __M2I_MIRANDA_DB_H

#include <stdio.h>
#include "types.h"

//all offsets are relative to the start of the file
//offsets are 0 if there is nothing in the chain or this is the last in the
//chain

/* tree diagram

DBHeader
 |-->end of file (plain offset)
 |-->first contact (DBContact)
 |   |-->next contact (DBContact)
 |   |   \--> ...
 |   |-->first settings (DBContactSettings)
 |   |    |-->next settings (DBContactSettings)
 |   |   |   \--> ...
 |   |   \-->module name (DBModuleName)
 |   \-->first/last/firstunread event
 |-->user contact (DBContact)
 |   |-->next contact=NULL
 |   |-->first settings   as above
 |   \-->first/last/firstunread event as above
 \-->first module name (DBModuleName)
     \-->next module name (DBModuleName)
        \--> ...
*/

#define DB_THIS_VERSION          0x00000700u

#pragma pack (push, 1)
struct DBHeader {
  BYTE signature[16];      // 'Miranda ICQ DB',0,26
  DWORD version;         //as 4 bytes, ie 1.2.3.10=0x0102030a
                           //this version is 0x00000700
  DWORD ofsFileEnd;         //offset of the end of the database - place to write
                           //new structures
  DWORD slackSpace;         //a counter of the number of bytes that have been
                           //wasted so far due to deleting structures and/or
                     //re-making them at the end. We should compact when
                     //this gets above a threshold
  DWORD contactCount;      //number of contacts in the chain,excluding the user
  DWORD ofsFirstContact;   //offset to first struct DBContact in the chain
  DWORD ofsUser;         //offset to struct DBContact representing the user
  DWORD ofsFirstModuleName;   //offset to first struct DBModuleName in the chain
};

#define DBCONTACT_SIGNATURE   0x43DECADEu
struct DBContact {
  DWORD signature;
  DWORD ofsNext;          //offset to the next contact in the chain. zero if
                             //this is the 'user' contact or the last contact
                      //in the chain
  DWORD ofsFirstSettings;    //offset to the first DBContactSettings in the
                             //chain for this contact.
  DWORD eventCount;          //number of events in the chain for this contact
  DWORD ofsFirstEvent,ofsLastEvent;    //offsets to the first and last DBEvent in
                                     //the chain for this contact
  DWORD ofsFirstUnreadEvent; //offset to the first (chronological) unread event
                      //in the chain, 0 if all are read
  DWORD timestampFirstUnread; //timestamp of the event at ofsFirstUnreadEvent
};

#define DBMODULENAME_SIGNATURE  0x4DDECADEu
struct DBModuleName {
  DWORD signature;
  DWORD ofsNext;      //offset to the next module name in the chain
  BYTE cbName;         //number of characters in this module name
  char name[0];         //name, no nul terminator
};

#define DBCONTACTSETTINGS_SIGNATURE  0x53DECADEu
struct DBContactSettings {
  DWORD signature;
  DWORD ofsNext;       //offset to the next contactsettings in the chain
  DWORD ofsModuleName;    //offset to the DBModuleName of the owner of these
                         //settings
  DWORD cbBlob;          //size of the blob in bytes. May be larger than the
                         //actual size for reducing the number of moves
                   //required using granularity in resizing
  BYTE blob[0];          //the blob. a back-to-back sequence of DBSetting
                         //structs, the last has cbName=0
};

/*   not a valid structure, content is figured out on the fly
struct DBSetting {
  BYTE cbName;         //number of bytes in the name of this setting
                        //this =0 marks the end
  char szName[...];      //setting name, excluding nul
  BYTE dataType;      //type of data. see m_database.h, db/contact/getsetting
  union {            //a load of types of data, length is defined by dataType
    BYTE bVal; WORD wVal; DWORD dVal;
   struct {
     WORD cbString;
     char szVal[...];     //excludes nul terminator
   };
   struct {
     WORD cbBlob;
     BYTE blobVal[...];
   };
  };
};
*/
//DBVARIANT: used by db/contact/getsetting and db/contact/writesetting
#define DBVT_DELETED 0    //this setting just got deleted, no other values are valid
#define DBVT_BYTE   1     //bVal and cVal are valid
#define DBVT_WORD   2     //wVal and sVal are valid
#define DBVT_DWORD  4     //dVal and lVal are valid
#define DBVT_ASCIIZ 255     //pszVal is valid
#define DBVT_BLOB   254     //cpbVal and pbVal are valid
#define DBVTF_VARIABLELENGTH  0x80
typedef struct {
   char *name;
   BYTE type;
   union {
      BYTE bVal; char cVal;
      WORD wVal; short sVal;
      DWORD dVal; long lVal;
      struct {
         char *pszVal;
         WORD cchVal;   //only used for db/contact/getsettingstatic
      };
      struct {
         WORD  cbVal;
         BYTE *pbVal;
      };
   };
} DBVARIANT;

#define DBEVENT_SIGNATURE  0x45DECADEu
struct DBEvent {
  DWORD signature;
  DWORD ofsPrev,ofsNext;    //offset to the previous and next events in the
                             //chain. Chain is sorted chronologically
  DWORD ofsModuleName;       //offset to a DBModuleName struct of the name of
                             //the owner of this event
  DWORD timestamp;  //seconds since 00:00, 01/01/1970. Gives us times until
                     //2106 unless you use the standard C library which is
                 //signed and can only do until 2038. In GMT.
  DWORD flags;             //see m_database.h, db/event/add
  WORD eventType;          //module-defined event type
  DWORD cbBlob;             //number of bytes in the blob
  BYTE blob[0];             //the blob. module-defined formatting
};

#define DBEF_FIRST    1    //this is the first event in the chain;
                           //internal only: *do not* use this flag
#define DBEF_SENT     2    //this event was sent by the user. If not set this
                           //event was received.
#define DBEF_READ     4    //event has been read by the user. It does not need
                           //to be processed any more except for history.
#define EVENTTYPE_MESSAGE   0
#define EVENTTYPE_URL       1
#define EVENTTYPE_ADDED         1000  //v0.1.1.0+: these used to be module-
#define EVENTTYPE_AUTHREQUEST   1001  //specific codes, hence the module-
#define EVENTTYPE_FILE          1002  //specific limit has been raised to 2000

#pragma pack (pop)

/** Verify that passed header is actually belongs to Miranda database.
*   @param header Header filled with information from db file.
*   @return @c true if header is valid, @c false otherwise.
*/
int ValidateDBHeader( DBHeader *header );

/** Read single contact entry from file fp into contact.
* @param filename Name of open file passed in fp (used for error reporting).
* @param fp File pointer of open database file.
* @param offset Offset in file to read contact entry from.
* @param contact Pointer to DBContact structure filled with information
*        upon return.
* @return @c true if read ok, dies breathless otherwise.
*/
int ReadContact( char *filename, FILE *fp, DWORD offset, DBContact *contact );

/** Read contact settings header.
* @param filename Name of open file passed in fp (used for error reporting).
* @param fp File pointer of open database file.
* @param offset Offset in file to read settings header from.
* @param settings Settings header filled upon return.
* @return @c true if read ok, dies breathless otherwise.
*/
int ReadContactSettings( char *filename, FILE *fp, DWORD offset, DBContactSettings *settings );

/** Read single setting from blob at offset.
* @param filename Name of open file passed in fp (used for error reporting).
* @param fp File pointer of open database file.
* @param offset Offset in file to read setting from.
* @param setting A setting filled with info upon return.
*                Necessary fields are allocated.
* @param bytesleft Number of bytes left in this blob.
*                  Decreased by number of bytes read on each call.
* @return Number of bytes read.
*/
DWORD ReadSingleContactSetting( char *filename, FILE *fp, DWORD offset, DBVARIANT *setting, DWORD *bytesleft );

/** Return textual name of contact setting.
* @param setting Setting to return type name for.
* @return Type name pointer to static string.
*/
const char *ContactSettingType(DBVARIANT *setting);

/** Get module name from module descriptor at offset.
* @param filename Name of open file passed in fp (used for error reporting).
* @param fp File pointer of open database file.
* @param offset Offset in file to read module descriptor from.
* @return Module name allocated on the heap. Program should delete it after use.
*/
char *ReadModuleName( char *filename, FILE *fp, DWORD offset );

/** Read single event at offset.
* @param filename Name of open file passed in fp (used for error reporting).
* @param fp File pointer of open database file.
* @param offset Offset in file to read module descriptor from.
* @param event Event structure filled with info upon return.
* @return Event blob data.
*/
BYTE *ReadContactEvent( char *filename, FILE *fp, DWORD offset, DBEvent *event );

/** Return textual name of contact event.
* @param event Event to return type name for.
* @return Type name pointer to static string.
*/
const char *ContactEventType(DBEvent *event);

#endif /* __M2I_MIRANDA_DB_H */
